#[derive(Debug, PartialEq)]
pub struct Page {
    id: String,
    body: String,
    footer: Footer,
}

#[derive(Debug, PartialEq)]
pub enum Footer {
    Ending,
    Goto(String),
    Choices(Vec<Choice>)
}

#[derive(Debug, PartialEq)]
pub struct Choice {
    option: u32,
    flavor: String,
    redirect: String,
    stat_check: Option<StatCheck>,
    stat_change: Option<StatChange>
}

#[derive(Debug, PartialEq)]
pub struct StatCheck {
    stat: String,
    value: i32,
    rel: char,
}

#[derive(Debug, PartialEq)]
pub struct StatChange {
    stat: String,
    addend: i32,
}


peg::parser! {
    grammar storybook_parser() for str {
        pub rule story() -> Vec<Page>
            = blankline()* ps:page()+ eof() { ps }
        rule page() -> Page
            = h:header() b:body() f:footer() blankline()* {
                Page {
                    id: h,
                    body: b,
                    footer: f
                }
            }
        rule header() -> String
            = i:id() __ blankline()* { i }
        rule id() -> String
            = i:$(['A'..='Z'|'_']+) _ { String::from(i) }
        rule body() -> String
            = b:$((!footer() [_])+) { String::from(b.trim()) }
        rule footer() -> Footer
            = "THE END" __ { Footer::Ending } /
              "GOTO" _ i:id() __ { Footer::Goto(i) } /
              c:choice()+ { Footer::Choices(c) }
        rule choice() -> Choice
            = o:$(['0'..='9']+) ")" _ f:$((!redirect() [_])+) r:redirect() __ {
                let (redirect, stat_change, stat_check) = r;
                Choice {
                    option: o.parse().unwrap(),
                    flavor: String::from(f.trim()),
                    redirect,
                    stat_change,
                    stat_check
                }
            }
        rule redirect() -> (String, Option<StatChange>, Option<StatCheck>)
            = ck:stat_check()? "[" i:id() "]" _ cg:stat_change()? { (i, cg, ck) }
        rule stat_change() -> StatChange
            = "(" a:$(['+'|'-'] ['0'..='9']+) _ n:stat_name() ")" _ { StatChange {
                addend: a.parse().unwrap(),
                stat: n
            }}
        rule stat_check() -> StatCheck
            = "<" n:stat_name() _ v:$(['0'..='9']+) r:['+'|'-'] ">" _ { StatCheck {
                stat: n,
                value: v.parse().unwrap(),
                rel: r
            }}
        rule stat_name() -> String
            = n:$(['a'..='z'|'A'..='Z']+) { String::from(n) }
        rule blankline()
            = _ __
        rule _
            = [' '|'\t']*
        rule __
            = "\r\n" / "\r" / "\n"
        rule eof()
            = ![_]
    }
}

pub fn parse_story(string: &str) -> Result<Vec<Page>,
                                           peg::error::ParseError<
                                                   peg::str::LineCol>> {
    storybook_parser::story(string)
}

#[cfg(test)]
mod tests {
    use super::*;
    const STORY: &str = r"
START

One day, Dane took a really big shit.

GOTO DIALOG

DIALOG

Dane: 'Wow, that was a really big shit!'

1) Die [DIE]
2) Don't die <Strength 100+> [DONTDIE] (-1 Honor)

DIE

Then he died.

Fuck you, Dane!

THE END
";
    #[test]
    fn storybook_parser_test() {
        assert_eq!(
            parse_story(STORY),
            Ok(vec![
                Page {
                    id: String::from("START"),
                    body: String::from("One day, Dane took a really big shit."),
                    footer: Footer::Goto(String::from("DIALOG"))
                },
                Page {
                    id: String::from("DIALOG"),
                    body: String::from("Dane: 'Wow, that was a really big shit!'"),
                    footer: Footer::Choices(vec![
                        Choice {
                            option: 1,
                            flavor: String::from("Die"),
                            redirect: String::from("DIE"),
                            stat_check: None,
                            stat_change: None,
                        },
                        Choice {
                            option: 2,
                            flavor: String::from("Don't die"),
                            redirect: String::from("DONTDIE"),
                            stat_change: Some(StatChange {
                                stat: String::from("Honor"),
                                addend: -1,
                            }),
                            stat_check: Some(StatCheck {
                                stat: String::from("Strength"),
                                value: 100,
                                rel: '+',
                            })
                        }
                    ])
                },
                Page {
                    id: String::from("DIE"),
                    body: String::from("Then he died.\n\nFuck you, Dane!"),
                    footer: Footer::Ending,
                }
            ])
        );
    }
}
